@@ -74,6 +74,11 @@ NATIVE_RETURN_POS = {
     "lookup": 1
 }
 
+NATIVE_SUB_RETURN_POS = {
+    "ucontrol.getBlock": 2,
+    "ucontrol.within": 3
+}
+
 NATIVE_CONST_INPUTS = {
     "draw": [0],
     "control": [0],
@@ -126,7 +131,7 @@ class Generator:
         self.func_stack = []
         self.loop_stack = []
         self.no_generate_tmp = False
-        self.var_list = set(self._generate_var_list(node))
+        self.var_list = set(self._generate_var_list(node) + ["true", "false", "null"])
 
         funcs = []
         for v in self.var_list:
@@ -136,8 +141,6 @@ class Generator:
                     gen_error(None, f"Redefinition of function \"{spl[0]}\"")
                 funcs.append(spl[0])
 
-        print(self.var_list)
-
         code = self._code_node_join(node)
 
         return self.postprocess(self.optimize(code, optimize_options))
@@ -243,6 +246,7 @@ class Generator:
             elif GEN_REGEXES["TCOMP"].fullmatch(ln):
                 spl = ln.split(" ", 4)
                 cond = spl[1]
+                name = spl[2]
                 op1 = spl[3]
                 op2 = spl[4]
 
@@ -439,17 +443,20 @@ class Generator:
 
             if GEN_REGEXES["IATTR"].fullmatch(node.value):
                 spl = node.value.split(".")
-
-                if spl[0] not in self.var_list:
-                    gen_undefined_error(node, spl[0])
                 
                 return f"sensor {var} {spl[0]} @{spl[1]}", var
             
-            if type(node.value) == str and not (node.value.startswith("\"") and node.value.endswith("\"")):
-                if GEN_REGEXES["VARU"].fullmatch(node.value) and node.value not in self.var_list:
+            if type(node.value) == str and not (node.value.startswith("\"") and node.value.endswith("\"")) and node.value != "_" and node.value[-1] not in "0123456789":
+                if GEN_REGEXES["VARU"].fullmatch(node.value) and node.value not in self.var_list and node.value not in functions.builtin:
                     gen_undefined_error(node, node.value)
 
             return f"set {var} {node.value}" if not self.no_generate_tmp else "", var
+        elif t == IndexNode:
+            var = self.get_tmp_var()
+
+            tmp, var_ = self._generate(node.index)
+            
+            return f"{tmp}\nread {var} {node.var} {var_}", var
         elif t == AtomNode:
             return self._generate(node.value)
         elif t == AssignmentNode:
@@ -469,6 +476,24 @@ class Generator:
                 
                 tmp, var = self._generate(node.right)
                 return f"{tmp}\nop {op} {node.left} {node.left} {var}"
+        elif t == IndexAssignNode:
+            if node.atype == "=":
+                itmp, ivar = self._generate(node.index)
+                vtmp, vvar = self._generate(node.val)
+
+                return f"{itmp}\n{vtmp}\nwrite {vvar} {node.var} {ivar}"
+            elif node.atype in ["+=", "-=", "*=", "/="]:
+                itmp, ivar = self._generate(node.index)
+                vtmp, vvar = self._generate(node.val)
+
+                at = node.atype
+                op = "add" if at == "+=" else "sub" if at == "-=" else "mul" if at == "*=" else "div" if at == "/=" else ""
+                if op == "":
+                    gen_error(node, f"Invalid operator: \"{at}\"")
+                
+                tval = self.get_tmp_var()
+
+                return f"{itmp}\n{vtmp}\nread {tval} {node.var} {ivar}\nop {op} {tval} {tval} {vvar}\nwrite {tval} {node.var} {ivar}"
         elif t == ExpressionNode:
             tmp, var = self._generate(node.left)
 
@@ -559,8 +584,6 @@ class Generator:
                     if not found:
                         gen_error(node, f"Undefined function \"{node.function}\"")
 
-                print(node.function, is_native, is_builtin)
-
                 tmp = ""
                 params = []
                 for i, arg in enumerate(node.params):
@@ -612,6 +635,27 @@ class Generator:
                 tmp, var = self._generate(node.function)
 
                 return tmp, var
+        elif t == SubCallNode:
+            spl = node.function.split(".")
+            req = functions.native_params[spl[0]]
+
+            tmp = ""
+            params = []
+            for i, arg in enumerate(node.params):
+                tmp_, var = self._generate(arg)
+
+                tmp += f"{tmp_}\n"
+                params.append(var)
+            
+            while len(params) < req - 1:
+                params.append("_")
+            
+            retpos = NATIVE_SUB_RETURN_POS.get(node.function, -1)
+            retvar = "null"
+            if retpos != -1:
+                retvar = params[retpos]
+            
+            return f"{tmp}{spl[0]} {spl[1]} {' '.join(params)}", retvar
         elif t == IfNode:
             tmp2, var = self._generate(node.condition)
 
@@ -638,7 +682,7 @@ class Generator:
             l_v = self.get_tmp_label()
             l_e = self.get_tmp_label()
 
-            self.loop_stack.append((l_v, l_e))
+            self.loop_stack.append((l_v, l_e, Node()))
 
             tmp = f"<{l_v}\n{tmp2}\n>{l_e} !{var}\n"
             tmp += self._code_node_join(node)
@@ -661,7 +705,7 @@ class Generator:
             l_v = self.get_tmp_label()
             l_e = self.get_tmp_label()
 
-            self.loop_stack.append((l_v, l_e))
+            self.loop_stack.append((l_v, l_e, node.action))
 
             tmp = f"{itmp}\n<{l_v}\n{ctmp}\n>{l_e} !{cvar}\n"
             tmp += self._code_node_join(node)
@@ -735,13 +779,13 @@ class Generator:
             if node.action == "break":
                 return f">{loop[1]}"
             elif node.action == "continue":
-                return f">{loop[0]}"
+                return f"{self._generate(loop[2])}\n>{loop[0]}"
             
             raise RuntimeError(f"Invalid AST")
         elif t == ExternNode:
             return ""
         
-        gen_error(node, "Unknown node")
+        gen_error(node, f"Unknown node ({node})")
     
     def _var_list_join(self, lists: list) -> list:
         tmp = []
@@ -756,10 +800,14 @@ class Generator:
             return self._var_list_join([self._generate_var_list(n) for n in node.code])
         elif t == ValueNode:
             return [node.value] if type(node.value) == str and not (node.value.startswith("\"") and node.value.endswith("\"")) else []
+        elif t == IndexNode:
+            return self._generate_var_list(node.index)
         elif t == AtomNode:
             return self._generate_var_list(node.value)
         elif t == AssignmentNode:
             return [node.left] if node.atype == "=" else [] + self._generate_var_list(node.right)
+        elif t == IndexAssignNode:
+            return []
         elif t in [ExpressionNode, CompExpressionNode, ArithExpNode, TermNode]:
             tmp = []
             tmp += self._generate_var_list(node.left)
@@ -779,6 +827,11 @@ class Generator:
                 return [f"__f_{node.function}_retv"] + tmp
             else:
                 return self._generate_var_list(node.function)
+        elif t == SubCallNode:
+            if node.function in NATIVE_SUB_RETURN_POS:
+                return self._generate_var_list(node.params[NATIVE_SUB_RETURN_POS[node.function]])
+            
+            return []
         elif t in [IfNode, WhileNode, ForNode, RepeatNode, FunctionNode]:
             tmp = self._var_list_join([self._generate_var_list(n) for n in node.code])
             if t in [IfNode, WhileNode, ForNode]:
@@ -801,7 +854,7 @@ class Generator:
         elif t == ExternNode:
             return [node.name]
 
-        gen_error(node, "Unknown node")
+        gen_error(node, f"Unknown node ({node})")
 
     def get_tmp_var(self) -> str:
         self.tmpv += 1
